(* Boulder Dash idea by Piter Liepa *)
(* Original BK ver by A.V.Melentiev *)
(* IBM PC version by Wladimir Mutel *)
(* Portable version by Oleg N. Cher *)

MODULE Grax; (* nonportable (i80x86) *)
(* Version for CGA 320x200,4 colors *)
IMPORT
  SYSTEM, Sys := Lowlevel, Settings,
  Timer, Rsrc := Resource, Keyboard;

CONST
  UseKeysColor  = TRUE;
  KeyPalette    = Keyboard.Key1; (* Press to change palette    *)
  KeyBkColor    = Keyboard.Key2; (* Press to change back color *)

  Black*        =  0;
  Blue*         =  1;
  Green*        =  2;
  Cyan*         =  3;
  Red*          =  4;
  Magenta*      =  5;
  Brown*        =  6;
  LightGray*    =  7;
  DarkGray*     =  8;
  LightBlue*    =  9;
  LightGreen*   = 10;
  LightCyan*    = 11;
  LightRed*     = 12;
  LightMagenta* = 13;
  Yellow*       = 14;
  White*        = 15;
TYPE
  COORDS*   = SHORTINT;
  SPRITE*   = SHORTINT;
  COLOR     = SHORTINT;
  PALETTE   = SHORTINT;
  TEXTCOLOR = SHORTINT;
VAR
  terminate*: BOOLEAN;
  TerminateProc*: PROCEDURE;
  pos, line: COORDS;
  color: TEXTCOLOR;
  bkcolor, palette: COLOR;

PROCEDURE PutSprite* (x, y: COORDS; spr: SPRITE);
  (* Draw CGA sprite 16x12 pixels, 4 colors *)
VAR
  sprAdr: Sys.ADDRESS;
BEGIN
  IF (x >= 0) & (x <= 30) & (y >= 0) & (y <= 30) THEN
    SYSTEM.CODE(
      8BH,76H,0AH,      (* mov si,word ptr[bp+10];x *)
      8BH,7EH,08H,      (* mov di,word ptr[bp+8] ;y *)
      03H,0F6H,         (* add si,si *)
      8BH,0CFH,         (* mov cx,di *)
      03H,0FFH,         (* add di,di *)
      03H,0FFH,         (* add di,di *)
      03H,0FFH,         (* add di,di *)
      03H,0FFH,         (* add di,di *)
      02BH,0F9H,        (* sub di,cx *)
      03H,0FFH,         (* add di,di *)
      03H,0FFH,         (* add di,di *)
      03H,0FFH,         (* add di,di *)
      03H,0FFH,         (* add di,di *)
      03H,0FEH,         (* add di,si *)
      81H,0C7H,0A8H,00H (* add di,0A8h *)
    );
    sprAdr := spr * Rsrc.SpriteSize + Rsrc.Sprites;
    SYSTEM.CODE(
      1EH,              (* push ds *)
      8BH,0F0H,         (* mov si,ax ;sprAdr *)
      8EH,0DAH,         (* mov ds,dx ;sprSeg *)
      0B8H,00H,0B8H,    (* mov ax,0B800h *)
      8EH,0C0H,         (* mov es,ax *)
      0FCH,             (* cld *)
      0B9H,06H,00H,     (* mov cx,6  *)
               (* drwline: *)
      0A5H,             (* movsw *)
      0A5H,             (* movsw *)
      81H,0C7H,0FCH,1FH,(* add di,1FFCh *)
      0A5H,             (* movsw *)
      0A5H,             (* movsw *)
      81H,0EFH,0B4H,1FH,(* sub di,1FB4h *)
      0E2H,0F2H,        (* loop drwline *)
      1FH               (* pop ds *)
    )
  END
END PutSprite;

PROCEDURE ClrScr* ;
BEGIN
  pos := 0; line := 0;
  SYSTEM.CODE(
    0B9H,11H,00H,      (* mov cx,11h     *)
    33H,0DBH,          (* xor bx,bx      *)
      (* clr_chunk: *)
    51H,53H);          (* push cx,bx     *)
  Timer.RunDelay;
  SYSTEM.CODE(
    0B8H,00H,0B8H,     (* mov ax,0B800h  *)
    8EH,0C0H,          (* mov es,ax      *)
    5BH,               (* pop bx         *)
      (* screen:    *)
    26H,88H,07H,       (* mov es:[bx],al *)
    83H,0C3H,11H,      (* add bx,11h     *)
    81H,0FBH,00H,40H,  (* cmp bx,4000h   *)
    72H,0F4H,          (* jb screen      *)
    81H,0E3H,0FFH,3FH, (* and bx,3FFFh   *)
    53H);              (* push bx        *)
  Timer.WaitDelay(1);
  SYSTEM.CODE(
    5BH,59H,           (* pop bx,cx      *)
    0E2H,0D5H)         (* loop clr_chunk *)
END ClrScr;

PROCEDURE CopyLine (to, from: INTEGER);
(* ()  from  to*)
BEGIN
  IF (to # from) & (to < 192) & (from < 192) THEN
    SYSTEM.CODE(
      8BH,7EH,06H,      (* mov di,word ptr[bp+6];to   *)
      8BH,76H,04H,      (* mov si,word ptr[bp+4];from *)
      83H,0C7H,04H,     (* add di,4    *)
      83H,0C6H,04H,     (* add si,4    *)
      8BH,0C7H,         (* mov ax,di   *)
      25H,0FEH,00H,     (* and ax,0FEh *)
      03H,0C0H,         (* add ax,ax   *)
      8BH,0C8H,         (* mov cx,ax   *)
      03H,0C0H,         (* add ax,ax   *)
      03H,0C0H,         (* add ax,ax   *)
      03H,0C1H,         (* add ax,cx   *)
      03H,0C0H,         (* add ax,ax   *)
      03H,0C0H,         (* add ax,ax   *)
      83H,0E7H,01H,     (* and di,1    *)
      0D1H,0CFH,        (* ror di,1    *)
      0D1H,0CFH,        (* ror di,1    *)
      0D1H,0CFH,        (* ror di,1    *)
      03H,0F8H,         (* add di,ax   *)
      8BH,0C6H,         (* mov ax,si   *)
      25H,0FEH,00H,     (* and ax,0FEh *)
      03H,0C0H,         (* add ax,ax   *)
      8BH,0C8H,         (* mov cx,ax   *)
      03H,0C0H,         (* add ax,ax   *)
      03H,0C0H,         (* add ax,ax   *)
      03H,0C1H,         (* add ax,cx   *)
      03H,0C0H,         (* add ax,ax   *)
      03H,0C0H,         (* add ax,ax   *)
      83H,0E6H,01H,     (* and si,1    *)
      0D1H,0CEH,        (* ror si,1    *)
      0D1H,0CEH,        (* ror si,1    *)
      0D1H,0CEH,        (* ror si,1    *)
      03H,0F0H,         (* add si,ax   *)
      0B8H,00H,0B8H,    (* mov ax,0B800h  *)
      8EH,0C0H,         (* mov es,ax   *)
      0FCH,             (* cld         *)
      0BBH,02H,00H,     (* mov bx,2    *)
      0B9H,28H,00H,     (* mov cx,28h  *)
               (* copyword: *)
      26H,8BH,04H,      (* mov ax,es:[si] *)
      0ABH,             (* stosw       *)
      03H,0F3H,         (* add si,bx   *)
      0E2H,0F8H         (* loop copyword  *)
    )
  END
END CopyLine;

PROCEDURE ClearLine (line: INTEGER);
(*     *)
BEGIN
  IF line < 192 THEN
    SYSTEM.CODE(
      8BH,7EH,04H,      (* mov di,word ptr[bp+4];line *)
      83H,0C7H,04H,     (* add di,4    *)
      8BH,0C7H,         (* mov ax,di   *)
      25H,0FEH,00H,     (* and ax,0FEh *)
      03H,0C0H,         (* add ax,ax   *)
      8BH,0C8H,         (* mov cx,ax   *)
      03H,0C0H,         (* add ax,ax   *)
      03H,0C0H,         (* add ax,ax   *)
      03H,0C1H,         (* add ax,cx   *)
      03H,0C0H,         (* add ax,ax   *)
      03H,0C0H,         (* add ax,ax   *)
      83H,0E7H,01H,     (* and di,1    *)
      0D1H,0CFH,        (* ror di,1    *)
      0D1H,0CFH,        (* ror di,1    *)
      0D1H,0CFH,        (* ror di,1    *)
      03H,0F8H,         (* add di,ax   *)
      0B8H,00H,0B8H,    (* mov ax,0B800h  *)
      8EH,0C0H,         (* mov es,ax   *)
      0FCH,             (* cld         *)
      33H,0C0H,         (* xor ax,ax   *)
      0B9H,28H,00H,     (* mov cx,28h  *)
      0F3H,0ABH,        (* rep stosw   *)
    )
  END
END ClearLine;

PROCEDURE ScrollUp* (lines: INTEGER);
VAR
  media, line: INTEGER;
BEGIN
  IF lines # 0 THEN
    IF lines < 192 THEN
      media := 192 - lines;
      (*for (line = 0; line < media; line ++)*)
      line := 0;
      WHILE line < media DO
        CopyLine(line, line + lines);
        INC(line)
      END;
      WHILE line < 192 DO
        ClearLine(line);
        INC(line)
      END;
      RETURN
    END;
    ClrScr
  END
END ScrollUp;

PROCEDURE ScrollDown* (lines: INTEGER);
VAR
  line: INTEGER;
BEGIN
  IF lines # 0 THEN
    IF lines < 192 THEN
      (*for (line = 191; line >= media; line --)*)
      line := 191;
      WHILE line >= lines DO
        CopyLine(line, line - lines);
        DEC(line)
      END;
      WHILE line >= 0 DO
        ClearLine(line);
        DEC(line)
      END;
      RETURN
    END;
    ClrScr
  END
END ScrollDown;

PROCEDURE SetPalette (p: PALETTE);
  (* 0 - Background, Green, Red, and Brown/Yellow; *)
  (* 1 - Background, Cyan, Magenta, and White.     *)
BEGIN
  SYSTEM.CODE(
    8AH,5EH,04H, (* mov bl,byte ptr[bp+4] *)
    0B7H,01H,    (* mov bh,1   *)
    0B4H,0BH,    (* mov ah,0Bh *)
    0CDH,10H     (* int 10h    *)
  )
END SetPalette;

PROCEDURE SetBkColor (c: COLOR);
BEGIN
  SYSTEM.CODE(
    8AH,5EH,04H, (* mov bl,byte ptr[bp+4] *)
    0B7H,00H,    (* mov bh,0   *)
    0B4H,0BH,    (* mov ah,0Bh *)
    0CDH,10H     (* int 10h    *)
  )
END SetBkColor;

PROCEDURE SetCursorPos* (x, y: COORDS);
BEGIN
  IF x < 32 THEN pos := x END;
  IF y < 16 THEN line := y END
END SetCursorPos;

PROCEDURE SetColor* (c: TEXTCOLOR);
BEGIN
  color := c
END SetColor;

PROCEDURE WriteCharXY (x, y: COORDS; ch: CHAR; c: TEXTCOLOR);
  (* Draw CGA font 8x12 pixels *)
VAR
  fontAdr: Sys.ADDRESS;
BEGIN
  SYSTEM.CODE(
    8BH,76H,0AH,      (* mov si,word ptr[bp+10];x *)
    8BH,7EH,08H,      (* mov di,word ptr[bp+8] ;y *)
    03H,0F6H,         (* add si,si *)
    8BH,0CFH,         (* mov cx,di *)
    03H,0FFH,         (* add di,di *)
    03H,0FFH,         (* add di,di *)
    03H,0FFH,         (* add di,di *)
    03H,0FFH,         (* add di,di *)
    02BH,0F9H,        (* sub di,cx *)
    03H,0FFH,         (* add di,di *)
    03H,0FFH,         (* add di,di *)
    03H,0FFH,         (* add di,di *)
    03H,0FFH,         (* add di,di *)
    03H,0FFH,         (* add di,di *)
    03H,0FEH,         (* add di,si *)
    81H,0C7H,0A8H,00H (* add di,0A8h *)
  );
  fontAdr := ORD(ch) * Rsrc.CharSize + Rsrc.Font;
  SYSTEM.CODE(
    1EH,              (* push ds *)
    8BH,0F0H,         (* mov si,ax ;fontAdr *)
    8EH,0DAH,         (* mov ds,dx ;fontSeg *)
    0B8H,00H,0B8H,    (* mov ax,0B800h *)
    8EH,0C0H,         (* mov es,ax *)
    0FCH,             (* cld *)
    0B1H,06H,         (* mov cl,6  *)
            (* drwline: *)
    0ADH,             (* lodsw *)
    8BH,0D8H,         (* mov bx,ax *)
    0B5H,10H,         (* mov ch,16 *)
            (* unpkbit: *)
    03H,0C0H,         (* add ax,ax *)
    13H,0D2H,         (* adc dx,dx *)
    03H,0C0H,         (* add ax,ax *)
    13H,0D2H,         (* adc dx,dx *)
    03H,0DBH,         (* add bx,bx *)
    73H,03H,          (* jnc nopoint *)
    0BH,46H,04H,      (* or ax,word ptr[bp+4] ;color *)
            (* nopoint: *)
    0FEH,0CDH,        (* dec ch *)
    75H,0EDH,         (* jnz unpkbit *)
    0ABH,             (* stosw *)
    81H,0C7H,0FEH,1FH,(* add di,1FFEh *)
    8BH,0C2H,         (* mov ax,dx *)
    0ABH,             (* stosw *)
    81H,0EFH,0B2H,1FH,(* sub di,1FB2h *)
    0E2H,0DAH,        (* loop drwline *)
    1FH               (* pop ds *)
  )
END WriteCharXY;

PROCEDURE WriteChar* (ch: CHAR);
BEGIN
  WriteCharXY (pos, line, ch, color);
  INC (pos);
  IF pos = 32 THEN (* New line *)
    INC (line);
    pos := 0;
    IF line = 16 THEN
      line := 15;
      ScrollUp(12)
    END
  END
END WriteChar;

PROCEDURE WriteString* (str: ARRAY OF CHAR);
VAR
  i: INTEGER;
BEGIN
  i := 0;
  WHILE str[i] # 00X DO
    WriteChar(str[i]); INC(i)
  END
END WriteString;

PROCEDURE Ln* ;
BEGIN
  IF line # 15 THEN
    INC(line)
  ELSE
    ScrollUp(12)
  END;
  pos := 0
END Ln;

PROCEDURE WriteInt* (i: INTEGER);
VAR
  b: ARRAY 6 OF CHAR;
  j: INTEGER;
BEGIN
  j := 5; b[5] := CHR(0);
  REPEAT
    IF i < 0 THEN
      WriteChar("-");
      IF i = 32768 THEN WriteString("32768"); RETURN END;
    i := -i;
    END;
    DEC(j);
    b[j] := CHR((i MOD 10) + ORD("0"));
    i := SHORT(LONG(i DIV 10))
  UNTIL j = 0;
  WriteString(b)
END WriteInt;

PROCEDURE KeyPressed* (key: Keyboard.KEY): BOOLEAN;
VAR
  i: INTEGER;
BEGIN
  IF key # Keyboard.AnyKey THEN
    IF UseKeysColor THEN
      IF Keyboard.Pressed(KeyPalette) THEN
        INC(palette); SetPalette(palette);
        Keyboard.Release(KeyPalette)
      END;
      IF Keyboard.Pressed(KeyBkColor) THEN
        INC(bkcolor); SetBkColor(bkcolor);
        Keyboard.Release(KeyBkColor)
      END
    END;
    RETURN Keyboard.Pressed(key)
  END;
  i := 128;
  REPEAT
    DEC(i);
    IF Keyboard.Pressed(CHR(i)) THEN
      IF UseKeysColor THEN
        CASE CHR(i) OF
          KeyPalette:
            INC(palette); SetPalette(palette);
            Keyboard.Release(KeyPalette)
        | KeyBkColor:
            INC(bkcolor); SetBkColor(bkcolor);
            Keyboard.Release(KeyBkColor)
        ELSE
          RETURN TRUE (* Key i pressed *)
        END
      ELSE
        RETURN TRUE (* Key i pressed *)
      END
    END
  UNTIL i = 0;
  RETURN FALSE (* No keys pressed *)
END KeyPressed;

PROCEDURE RunDelay* ;
BEGIN
  Timer.RunDelay
END RunDelay;

PROCEDURE WaitDelay* (ticks: Timer.TICKS);
BEGIN
  Timer.WaitDelay(ticks)
END WaitDelay;

PROCEDURE Close* ;
BEGIN
  Keyboard.Close;
  IF Settings.GraxWaitAtExit THEN SYSTEM.CODE(
    32H,0E4H,    (* xor ah,ah   *)
    0CDH,16H)    (* int 16h     *)
  END;
  SYSTEM.CODE(
    0B8H,03H,0H, (* mov ax,03h  *)
    0CDH,10H     (* int 10h     *)
  )
END Close;

BEGIN
  TerminateProc := Close;
  terminate := FALSE;
  color := 3;
  SYSTEM.CODE(
    0B8H,04H,0H, (* mov ax,04h  *)
    0CDH,10H     (* int 10h     *)
  );
  SetBkColor(Black); SetPalette(16);
  IF UseKeysColor THEN bkcolor := Black; palette := 16 END
END Grax.
